// Library     : DPMI
// Description : DOS Protected Mode Interface function library
// Module      : DPMI03.C
// Version     : 1.00 - 01/09/1994
// Written by  : David A. Mair
// -----------------------------------------------------------------------------
// Change summary --------------------------------------------------------------
// -----------------------------------------------------------------------------
//    Date   | Version | Comment and author
// -----------------------------------------------------------------------------
//           |         |
//           |         |
// -----------------------------------------------------------------------------


// Inline assembler is included in this module
#pragma inline

#include "dpmi.h"


#ifdef __DPMI32__
WORD SimulateRealModeInterrupt(BYTE byIntNum, REAL_MODE_REGS *pRMRegs, void *pStackData, WORD wStackDataWords)
#else
WORD SimulateRealModeInterrupt(BYTE byIntNum, REAL_MODE_REGS far *fpRMRegs, void far *fpStackData, WORD wStackDataWords)
#endif
{
	WORD wRetVal;
	WORD wSavedES;
	
		/* First build the data that must be on the stack */
#ifdef __DPMI32__
	asm {
		// Save the stack pointer in edx
		mov		edx, esp

		// Get the pointer to the data to be stacked
		mov		ebx, dword ptr [pStackData]

		// Get the number of words to push
		mov		cx, word ptr [wStackDataWords]

		// Point ebx to the end of the data to be stacked
		movzx	ecx, cx
		add		ebx, ecx
		add		ebx, ecx
	}

srmi32StackNextItem:
	asm {
		sub		ebx, 0x00000002
		push	word ptr [ebx]
		sub		cx, 1
		jnz		srmi32StackNextItem

		/* Now set up the DPMI call */
		mov		word ptr [wSavedES], es
		mov		ax, ds
		mov		es, ax
		mov		ax, 0x0300
		mov		bl, byte ptr [byIntNum]
		mov		bh, 0x00
		mov		cx, word ptr [wStackDataWords]
		mov		edi, dword ptr [pRMRegs]
		int		0x31

		mov		es, word ptr [wSavedES]
		mov		esp, edx

		jc		srmiSetRetVal

		mov		ax, 0x0000
	}
#else
	asm {
		// Save the stack pointer in dx
		mov		dx, sp

		// Get the pointer to the data to be stacked
		mov		word ptr [wSavedES], es
		les		bx, fpStackData

		// Get the number of words to push
		mov		cx, word ptr [wStackDataWords]

		// Fix up bx to point to the end of the buffer
		add		bx, cx
		add		bx, cx
	}

srmi16StackNextItem:
	asm {
		sub		bx, 0x0002
		push	word ptr es:[bx]
		sub		cx, 1
		jnz		srmi16StackNextItem

		/* Now set up the DPMI call */
		mov		ax, 0x0300
		mov		bl, byte ptr [byIntNum]
		mov		bh, 0x00
		mov		cx, word ptr [wStackDataWords]
		les		di, fpRMRegs
		int		0x31

		mov		es, word ptr [wSavedES]
		mov		sp, dx

		jc		srmiSetRetVal

		mov		ax, 0x0000
	}
#endif

srmiSetRetVal:
	asm {
		mov		word ptr [wRetVal], ax
	}

	return wRetVal;
}


#ifdef __DPMI32__
WORD SimulateRealModeFarCall(REAL_MODE_REGS *pRMRegs, void *pStackData, WORD wStackDataWords)
#else
WORD SimulateRealModeFarCall(REAL_MODE_REGS far *pRMRegs, void far *pStackData, WORD wStackDataWords)
#endif
{
	WORD wRetVal;
	WORD wSavedES;

#ifdef __DPMI32__	
	asm {
		// Save the stack pointer in edx
		mov		edx, esp

		// Get the pointer to the data to be stacked
		mov		ebx, dword ptr [pStackData]

		// Get the number of words to push
		mov		cx, word ptr [wStackDataWords]

		// Point ebx to the end of the data to be stacked
		movzx	ecx, cx
		add		ebx, ecx
		add		ebx, ecx
	}

srmfc32StackNextItem:
	asm {
		sub		ebx, 0x00000002
		push	word ptr [ebx]
		sub		cx, 1
		jnz		srmfc32StackNextItem

		/* Now set up the DPMI call */
		mov		word ptr [wSavedES], es
		mov		ax, ds
		mov		es, ax
		mov		ax, 0x0301
		mov		bh, 0x00
		mov		cx, word ptr [wStackDataWords]
		mov		edi, dword ptr [pRMRegs]
		int		0x31

		mov		es, word ptr [wSavedES]
		mov		esp, edx

		jc		srmfcSetRetVal

		mov		ax, 0x0000
	}
#else
	asm {
		// Save the stack pointer in dx
		mov		dx, sp

		// Get the pointer to the data to be stacked
		mov		word ptr [wSavedES], es
		les		bx, pStackData

		// Get the number of words to push
		mov		cx, word ptr [wStackDataWords]

		// Fix up bx to point to the end of the buffer
		add		bx, cx
		add		bx, cx
	}

srmfc16StackNextItem:
	asm {
		sub		bx, 0x0002
		push	word ptr es:[bx]
		sub		cx, 1
		jnz		srmfc16StackNextItem

		/* Now set up the DPMI call */
		mov		ax, 0x0301
		mov		bh, 0x00
		mov		cx, word ptr [wStackDataWords]
		les		di, pRMRegs
		int		0x31

		mov		es, word ptr [wSavedES]
		mov		sp, dx

		jc		srmfcSetRetVal

		mov		ax, 0x0000
	}
#endif

srmfcSetRetVal:
	asm {
		mov		word ptr [wRetVal], ax
	}

	return wRetVal;
}


#ifdef __DPMI32__
WORD SimulateRealModeFarCallToInterrupt(REAL_MODE_REGS *pRMRegs, void *pStackData, WORD wStackDataWords)
#else
WORD SimulateRealModeFarCallToInterrupt(REAL_MODE_REGS far *fpRMRegs, void far *fpStackData, WORD wStackDataWords)
#endif
{
	WORD wRetVal;
	WORD wSavedES;

#ifdef __DPMI32__	
	asm {
		// Save the stack pointer in edx
		mov		edx, esp

		// Get the pointer to the data to be stacked
		mov		ebx, dword ptr [pStackData]

		// Get the number of words to push
		mov		cx, word ptr [wStackDataWords]

		// Point ebx to the end of the data to be stacked
		movzx	ecx, cx
		add		ebx, ecx
		add		ebx, ecx
	}

srmfcti32StackNextItem:
	asm {
		sub		ebx, 0x00000002
		push	word ptr [ebx]
		sub		cx, 1
		jnz		srmfcti32StackNextItem

		/* Now set up the DPMI call */
		mov		word ptr [wSavedES], es
		mov		ax, ds
		mov		es, ax
		mov		ax, 0x0302
		mov		bh, 0x00
		mov		cx, word ptr [wStackDataWords]
		mov		edi, dword ptr [pRMRegs]
		int		0x31

		mov		es, word ptr [wSavedES]
		mov		esp, edx

		jc		srmfctiSetRetVal

		mov		ax, 0x0000
	}
#else
	asm {
		// Save the stack pointer in dx
		mov		dx, sp

		// Get the pointer to the data to be stacked
		mov		word ptr [wSavedES], es
		les		bx, fpStackData

		// Get the number of words to push
		mov		cx, word ptr [wStackDataWords]

		// Fix up bx to point to the end of the buffer
		add		bx, cx
		add		bx, cx
	}

srmfcti16StackNextItem:
	asm {
		sub		bx, 0x0002
		push	word ptr es:[bx]
		sub		cx, 1
		jnz		srmfcti16StackNextItem

		/* Now set up the DPMI call */
		mov		ax, 0x0302
		mov		bh, 0x00
		mov		cx, word ptr [wStackDataWords]
		les		di, fpRMRegs
		int		0x31

		mov		es, word ptr [wSavedES]
		mov		sp, dx

		jc		srmfctiSetRetVal

		mov		ax, 0x0000
	}
#endif

srmfctiSetRetVal:
	asm {
		mov		word ptr [wRetVal], ax
	}

	return wRetVal;
}


#ifdef __DPMI32__
WORD AllocateRealModeCallbackAddress(void *pCallbackProc, REAL_MODE_REGS *pRMRegs, REAL_MODE_FAR_POINTER *pCallbackAddress)
#else
WORD AllocateRealModeCallbackAddress(void far *fpCallbackProc, REAL_MODE_REGS far *fpRMRegs, REAL_MODE_FAR_POINTER far *fpCallbackAddress)
#endif
{
	WORD wRetVal;
	WORD wSavedDS;
	WORD wSavedES;

	asm {
		mov		word ptr [wSavedDS], ds
		mov		word ptr [wSavedES], es

#ifdef __DPMI32__		
		mov		es, word ptr [wSavedDS]
		mov		edi, dword ptr [pRMRegs]

		mov		ax, cs
		mov		ds, ax
		mov		esi, dword ptr [pCallbackProc]
#else
		les		di, fpRMRegs
		lds		si, fpCallbackProc
#endif

		mov		ax, 0x0303
		int		0x31

		mov		ds, word ptr [wSavedDS]
		mov		es, word ptr [wSavedES]
		
		jc		armcaSetRetVal

#ifdef __DPMI32__
		mov		ebx, dword ptr [pCallbackAddress]
		mov		word ptr [ebx], dx
		mov		word ptr [ebx + 2], cx
#else
		les		di, fpCallbackAddress
		mov		word ptr es:[di], dx
		mov		word ptr es:[di + 2], cx
		mov		es, word ptr [wSavedES]
#endif

		mov		ax, 0x0000
	}

armcaSetRetVal:
	asm {
		mov		word ptr [wRetVal], ax
	}

	return wRetVal;
}


#ifdef __DPMI32__
WORD FreeRealModeCallbackAddress(REAL_MODE_FAR_POINTER *pCallbackAddress)
#else
WORD FreeRealModeCallbackAddress(REAL_MODE_FAR_POINTER far *pCallbackAddress)
#endif
{
	WORD wRetVal;

	asm {
		mov		ax, 0x0304
#ifdef __DPMI32__
		mov		ebx, dword ptr [pCallbackAddress]
		mov		dx, word ptr [ebx]
		mov		cx, word ptr [ebx + 2]
#else
		push	es
		les		bx, pCallbackAddress
		mov		dx, word ptr es:[bx]
		mov		cx, word ptr es:[bx + 2]
		pop		es
#endif
		int		0x31
		
		jc		frmcaSetRetVal
		
		mov		ax, 0x0000
	}

frmcaSetRetVal:
	asm {
		mov		word ptr [wRetVal], ax
	}

	return wRetVal;
}


#ifdef __DPMI32__
WORD GetSaveRestoreAddresses(REAL_MODE_FAR_POINTER *pRealModeRoutine, PROT_MODE_FAR_POINTER_32 *pProtectedModeRoutine)
#else
WORD GetSaveRestoreAddresses(REAL_MODE_FAR_POINTER far *fpRealModeRoutine, PROT_MODE_FAR_POINTER_16 far *fpProtectedModeRoutine)
#endif
{
	WORD wStateBufferSize;

	asm {
		mov		ax, 0x0305
		int		0x31

		mov		word ptr [wStateBufferSize], ax

#ifdef __DPMI32__
		push	esi
		push	edi
		mov		edx, dword ptr [pRealModeRoutine]
		mov		word ptr [edx], cx
		mov		word ptr [edx + 2], bx

		mov		edx, dword ptr [pProtectedModeRoutine]
		mov		dword ptr [edx], edi
		mov		word ptr [edx + 4], si
		pop		edi
		pop		esi
#else
		push	es
		push	si
		push	di
		mov		dx, bx
		
		les		bx, fpRealModeRoutine
		mov		word ptr es:[bx], cx
		mov		word ptr es:[bx + 2], dx

		les		bx, fpProtectedModeRoutine
		mov		word ptr es:[bx], di
		mov		word ptr es:[bx + 2], si

		pop		di
		pop		si
		pop		es
#endif
	}

	return wStateBufferSize;
}


#ifdef __DPMI32__
void GetRawModeSwitchAddresses(REAL_MODE_FAR_POINTER *pRealToProtectedMode, PROT_MODE_FAR_POINTER_32 *pProtectedToRealMode)
#else
void GetRawModeSwitchAddresses(REAL_MODE_FAR_POINTER far *fpRealToProtectedMode, PROT_MODE_FAR_POINTER_16 far *fpProtectedToRealMode)
#endif
{
	asm {
		mov		ax, 0x0306
		int		0x31

#ifdef __DPMI32__
		push	esi
		push	edi
		mov		edx, dword ptr [pRealToProtectedMode]
		mov		word ptr [edx], cx
		mov		word ptr [edx + 2], bx

		mov		edx, dword ptr [pProtectedToRealMode]
		mov		dword ptr [edx], edi
		mov		word ptr [edx + 4], si
		pop		edi
		pop		esi
#else
		push	es
		push	si
		push	di
		mov		dx, bx
		
		les		bx, fpRealToProtectedMode
		mov		word ptr es:[bx], cx
		mov		word ptr es:[bx + 2], dx

		les		bx, fpProtectedToRealMode
		mov		word ptr es:[bx], di
		mov		word ptr es:[bx + 2], si

		pop		di
		pop		si
		pop		es
#endif
	}
}

